//! Solana support for Icepick.
//!
//! # Command Line Arguments
//!
//! The first thing you'll want is a wallet. For this, you can run the command
//! `icepick sol generate-wallet`. The output of the command will be a request to derive a key,
//! which Icepick will interpret upon subsequent invocations and replace with the derived key -
//! after appending the valid `bip44` path. You can then pipe the output of the command to
//! `icepick sol get-wallet-address`, which will take the derived key and format it as a Solana
//! address. In full, the commands look like:
//!
//! ```sh
//! icepick sol generate-wallet | icepick sol get-wallet-address
//! ```
//!
//! Next, you'll want to either airdrop some funds into the wallet if on devnet, or transfer some
//! funds using another source if on mainnet.
//!
//!
//! To transfer funds out of the wallet, you'll need a recent blockhash. The command
//! `icepick sol get-blockhash` can be used to get a recent blockhash. This blockhash must be used
//! within 150 blocks, or about 1 minute, or transactions won't be accepted.
//!
//! The blockhash is then used in the next command, `icepick sol transfer`, which also requires an
//! amount, a to-address, and a from-address. The output of this command can then be saved to
//! removable media, then transferred to an offline signer system where `icepick sol sign` is used
//! to sign the transaction, with the signed result also persisted to removable media. Once the
//! signed transaction is back on an online system, the transaction can be broadcasted using
//! `icepick sol broadcast`.
//!
//! ```sh
//! # On an online system
//! blockhash=$(icepick sol get-blockhash | jq -r .blob)
//! icepick sol transfer $amount $to_address $from_address $blockhash > sdcard/unsigned.json
//!
//! # On an offline system
//! icepick sol sign < sdcard/unsigned.json > sdcard/signed.json
//!
//! # On the online system, again
//! icepick sol broadcast < sdcard/signed.json
//! ```

use icepick_module::{
    help::{Argument, ArgumentType},
    Module,
};
use serde::{Deserialize, Serialize};
use solana_sdk::signer::Signer;
use std::str::FromStr;

// How does this not exist in solana_sdk.
const LAMPORTS_PER_SOL: u64 = 1_000_000_000;

#[derive(thiserror::Error, Debug)]
pub enum Error {}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub enum Cluster {
    Devnet,
    Testnet,
    Mainnet,
}

impl std::fmt::Display for Cluster {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Cluster::Devnet => f.write_str("devnet"),
            Cluster::Testnet => f.write_str("testnet"),
            Cluster::Mainnet => f.write_str("mainnet"),
        }
    }
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct GetBlockhash {
    cluster: Option<Cluster>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct GenerateWallet {
    account: Option<String>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct GetWalletAddress {}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct Transfer {
    amount: String,
    blockhash: String,
    to_address: String,
    from_account: Option<String>,
    from_address: String,
    fee: Option<String>,
    fee_payer: Option<String>,
    fee_payer_address: Option<String>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct Sign {}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct Broadcast {
    cluster: Option<Cluster>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "kebab-case")]
pub struct Request {
    // NOTE: Can't use the proper XPrv type from Keyfork because Solana's a big stinky
    // and adds in its own derivation constructs that cause type conflicts.
    derived_keys: Option<Vec<[u8; 32]>>,

    // NOTE: This is an opaque type that can be deserialized inside an Operation
    blob: Option<serde_json::Value>,

    #[serde(flatten)]
    operation: Operation,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(tag = "operation", content = "values", rename_all = "kebab-case")]
pub enum Operation {
    GetBlockhash(GetBlockhash),
    GenerateWallet(GenerateWallet),
    GetWalletAddress(GetWalletAddress),
    Transfer(Transfer),
    Sign(Sign),
    Broadcast(Broadcast),
}

pub struct Solana;

impl Solana {
    fn keypair_from_bytes(given_bytes: [u8; 32]) -> solana_sdk::signer::keypair::Keypair {
        use ed25519_dalek::{PublicKey, SecretKey};
        let secret_key = SecretKey::from_bytes(&given_bytes).expect("key should be 32 bytes");
        let mut bytes = [0u8; 64];
        bytes[..32].clone_from_slice(&given_bytes);
        bytes[32..].clone_from_slice(PublicKey::from(&secret_key).as_bytes());
        solana_sdk::signer::keypair::Keypair::from_bytes(&bytes)
            .expect("solana sdk should expect 64 bytes")
    }
}

impl Module for Solana {
    type Error = Error;

    type Request = Request;

    fn describe_operations() -> Vec<icepick_module::help::Operation> {
        let cluster = Argument {
            name: "cluster".to_string(),
            description: "The cluster to interact with (mainnet, testnet, devnet)".to_string(),
            r#type: ArgumentType::Optional,
        };
        let account = Argument {
            name: "from-account".to_string(),
            description: "The derivation account used for the transaction.".to_string(),
            r#type: ArgumentType::Optional,
        };
        let fee = Argument {
            name: "fee".to_string(),
            description: "A custom fee for the transaction".to_string(),
            r#type: ArgumentType::Optional,
        };
        let fee_payer_address = Argument {
            name: "fee-payer-address".to_string(),
            description: "The address used to pay the fee.".to_string(),
            r#type: ArgumentType::Optional,
        };
        let fee_payer = Argument {
            name: "fee-payer".to_string(),
            description: "The derivation account used to pay the fee.".to_string(),
            r#type: ArgumentType::Optional,
        };
        let blockhash = Argument {
            name: "blockhash".to_string(),
            description: "A recent blockhash".to_string(),
            r#type: ArgumentType::Required,
        };
        let from_address = Argument {
            name: "from-address".to_string(),
            description: concat!(
                "The address to send SOL from; will be used to verify ",
                "the derivation account."
            )
            .to_string(),
            r#type: ArgumentType::Required,
        };
        vec![
            icepick_module::help::Operation {
                name: "get-blockhash".to_string(),
                description: "Get the latest blockhash".to_string(),
                arguments: vec![cluster.clone()],
            },
            icepick_module::help::Operation {
                name: "generate-wallet".to_string(),
                description: "Generate the derivation index for a wallet.".to_string(),
                arguments: vec![Argument {
                    name: "account".to_string(),
                    description: "The derivation account used for generating the wallet."
                        .to_string(),
                    r#type: ArgumentType::Optional,
                }],
            },
            icepick_module::help::Operation {
                name: "get-wallet-address".to_string(),
                description: "Get the address for a given wallet.".to_string(),
                arguments: vec![],
            },
            icepick_module::help::Operation {
                name: "transfer".to_string(),
                description: "Transfer SOL from a Keyfork wallet to an external wallet."
                    .to_string(),
                arguments: vec![
                    Argument {
                        name: "amount".to_string(),
                        description: "The amount of SOL to transfer.".to_string(),
                        r#type: ArgumentType::Required,
                    },
                    account.clone(),
                    Argument {
                        name: "to-address".to_string(),
                        description: "The address to send SOL to.".to_string(),
                        r#type: ArgumentType::Required,
                    },
                    from_address.clone(),
                    blockhash.clone(),
                    fee.clone(),
                    fee_payer.clone(),
                    fee_payer_address.clone(),
                ],
            },
            icepick_module::help::Operation {
                name: "stake".to_string(),
                description: "Stake SOL to earn rewards.".to_string(),
                arguments: vec![
                    Argument {
                        name: "amount".to_string(),
                        description: "The amount of SOL to stake.".to_string(),
                        r#type: ArgumentType::Required,
                    },
                    account.clone(),
                    from_address.clone(),
                    blockhash.clone(),
                    fee.clone(),
                    fee_payer.clone(),
                    fee_payer_address.clone(),
                ],
            },
            icepick_module::help::Operation {
                name: "sign".to_string(),
                description: "Sign a previously-generated transaction.".to_string(),
                arguments: vec![],
            },
            icepick_module::help::Operation {
                name: "broadcast".to_string(),
                description: "Broadcast a signed transaction".to_string(),
                arguments: vec![cluster.clone()],
            },
        ]
    }

    fn handle_request(request: Self::Request) -> Result<serde_json::Value, Self::Error> {
        match request.operation {
            Operation::GetBlockhash(GetBlockhash { cluster }) => {
                let cluster = cluster.unwrap_or(Cluster::Mainnet);
                let cluster_url = format!("https://api.{cluster}.solana.com");
                let client = solana_rpc_client::rpc_client::RpcClient::new(cluster_url);
                let response = client.get_latest_blockhash().unwrap();
                Ok(serde_json::json!({
                    "blob": response.to_string(),
                }))
            }
            Operation::GenerateWallet(GenerateWallet { account }) => {
                let account = u32::from_str(account.as_deref().unwrap_or("0")).unwrap();
                Ok(serde_json::json!({
                    "blob": null,
                    "derivation-accounts": [(account | 1 << 31)],
                }))
            }
            Operation::GetWalletAddress(_) => {
                // NOTE: panics if doesn't exist
                let key = request.derived_keys.unwrap()[0];
                let keypair = Self::keypair_from_bytes(key);
                let pubkey = keypair.pubkey();
                Ok(serde_json::json!({
                    "blob": pubkey.to_string(),
                }))
            }
            Operation::Transfer(Transfer {
                amount,
                from_account,
                to_address,
                from_address,
                blockhash,
                fee: _,
                fee_payer,
                fee_payer_address,
            }) => {
                // TODO:
                // parse address for to_address

                let amount = f64::from_str(&amount).expect("float amount");
                let amount: u64 = (amount * LAMPORTS_PER_SOL as f64) as u64;

                use solana_sdk::pubkey::Pubkey;
                let to_pk = Pubkey::from_str(&to_address).unwrap();
                let from_pk = Pubkey::from_str(&from_address).unwrap();
                let payer_account_and_pk = {
                    // If a fee payer is given, a fee payer address must also be given, since the
                    // address must be known before signing the transaction.
                    match (&fee_payer, &fee_payer_address) {
                        (Some(payer), Some(address)) => {
                            // Use the provided account
                            Some((
                                u32::from_str(payer).unwrap(),
                                Pubkey::from_str(address).unwrap(),
                            ))
                        }
                        (None, None) => {
                            // Use the transaction account
                            None
                        }
                        _ => panic!("Invalid combination of fee_payer and fee_payer_address"),
                    }
                };
                let instruction =
                    solana_sdk::system_instruction::transfer(&from_pk, &to_pk, amount);
                let hash = solana_sdk::hash::Hash::from_str(&blockhash).unwrap();
                let message = solana_sdk::message::Message::new_with_blockhash(
                    &[instruction],
                    payer_account_and_pk.map(|v| v.1).as_ref(),
                    &hash,
                );
                let transaction = solana_sdk::transaction::Transaction::new_unsigned(message);
                // TODO: error handling from_str
                let from_account = from_account
                    .and_then(|a| u32::from_str(&a).ok())
                    .unwrap_or(0);
                let mut requested_accounts = vec![];
                requested_accounts.push(from_account | 1 << 31);
                if let Some((account, _)) = &payer_account_and_pk {
                    requested_accounts.push(*account | 1 << 31);
                }
                Ok(serde_json::json!({
                    "blob": transaction,
                    "derivation-accounts": requested_accounts,
                }))
            }
            Operation::Sign(_) => {
                let blob = request.blob.expect("passed in instruction blob");
                let mut transaction: solana_sdk::transaction::Transaction =
                    serde_json::from_value(blob).expect("valid message blob");
                let keys = request
                    .derived_keys
                    .unwrap_or_default()
                    .iter()
                    .map(|k| Self::keypair_from_bytes(*k))
                    .collect::<Vec<_>>();

                let hash = transaction.message.recent_blockhash;
                transaction
                    .try_sign(&keys, hash)
                    .expect("not enough keys provided");
                Ok(serde_json::json!({
                    "blob": transaction,
                }))
            }
            Operation::Broadcast(Broadcast { cluster }) => {
                let cluster = cluster.unwrap_or(Cluster::Mainnet);
                let cluster_url = format!("https://api.{cluster}.solana.com");

                let blob = request.blob.expect("passed in instruction blob");
                let transaction: solana_sdk::transaction::Transaction =
                    serde_json::from_value(blob).expect("valid message blob");
                transaction.verify().expect("invalid signatures");
                let client = solana_rpc_client::rpc_client::RpcClient::new(cluster_url);
                let _simulated_response = client.simulate_transaction(&transaction).unwrap();
                let response = client.send_and_confirm_transaction(&transaction);
                Ok(match response {
                    Ok(s) => {
                        serde_json::json!({
                            "blob": {
                                "status": "send_and_confirm",
                                "succcess": s.to_string(),
                            }
                        })
                    }
                    Err(e) => {
                        serde_json::json!({
                            "blob": {
                                "status": "send_and_confirm",
                                "error": e.to_string(),
                            }
                        })
                    }
                })
            }
        }
    }
}
